from .sagemaker_endpoint import SagemakerEndpoint
from .s3_hashmap import S3Hashmap
from .s3_dataframe import S3Dataframe
from .s3_dataframe_set import S3DataframeSet
from .s3_scikit import S3Scikit
from .s3_scikit_set import S3ScikitSet


def create_ds_models(**kwargs):
    return DsModels(**kwargs)


def create_ds_model(*, hdbscan=None, model, log):
    model_type = model["type"]
    if model_type == "sagemaker_endpoint":
        return SagemakerEndpoint(model=model, log=log)
    if model_type == "s3_hashmap":
        return S3Hashmap(model=model, log=log)
    if model_type == "s3_dataframe":
        return S3Dataframe(model=model, log=log)
    if model_type == "s3_dataframe_set":
        return S3DataframeSet(model=model, log=log)
    if model_type == "s3_scikit_set":
        return S3ScikitSet(hdbscan=hdbscan, model=model, log=log)
    if model_type == "s3_scikit":
        return S3Scikit(model=model, log=log)

    raise Exception(f"Unknown model type: {model_type}")


class DsModels:
    def __init__(self, *, hdbscan=None, models, log):
        self._models = models
        self._hdbscan = hdbscan
        self._log = log

    def get_ds_model(self, ds_model_name, /):
        model_data = self._pick_model_version(ds_model_name)
        log = self._log.bind(ds_model=ds_model_name)
        log.info("Selected model version", model_meta=model_data["model_name"])
        log.debug("Model data:", model_data=model_data)
        return create_ds_model(hdbscan=self._hdbscan, model=model_data, log=log)

    def _pick_model_version(self, model_name):
        versions = self._models.get(model_name)
        if versions is None:
            raise Exception(f"No model named: {model_name}")
        if len(versions) == 0:
            raise Exception(f"No versions for model: {model_name}")

        return versions[0]



# pylint: disable=missing-docstring
# pylint: disable=unused-import
# pylint: disable=unused-argument
import os
import itertools
import pytest
import pandas as pd
from .curator import TomeCuratorFs


# pylint: disable=invalid-name
ds_collection_root_path = "fixtures"
ds_type = "csds"
default_header_name = "header_tome"
sub_header_name = "only_good_headers"
new_tome_name = "round_end"
continued_tome_name = "round_end_continued"
pd.DataFrame()


def create_curator_instance(tmp_path):
    return TomeCuratorFs(
        default_header_name=default_header_name,
        ds_type=ds_type,
        tome_collection_root_path=tmp_path,
        ds_collection_root_path=ds_collection_root_path,
    )


def create_header(curator):
    return curator.create_header_tome()


def create_header_and_subheader(curator):
    create_header(curator)
    return curator.create_subheader_tome(
        sub_header_name,
        lambda df: df["key"]
        != "csds/2022/05/15/63cc7181-07c9-42fd-ade4-4eeb2cf4db6f/csds",
    )


def test_create_header_tome(tmp_path):
    tmp_path = str(tmp_path)
    curator = create_curator_instance(tmp_path)
    loader = create_header(curator)

    df = loader.get_dataframe()
    keyset = loader.get_keyset()
    manifest = loader.manifest

    assert len(df) == len(keyset)
    assert len(df) == 3
    assert isinstance(manifest, dict)


def test_create_subheader_tome(tmp_path):
    tmp_path = str(tmp_path)
    curator = create_curator_instance(tmp_path)
    loader = create_header_and_subheader(curator)

    df = loader.get_dataframe()
    keyset = loader.get_keyset()
    manifest = loader.manifest
    assert len(df) == len(keyset)
    assert len(df) == 2
    assert isinstance(manifest, dict)


def test_make_tome(tmp_path):
    tmp_path = str(tmp_path)
    curator = create_curator_instance(tmp_path)
    create_header_and_subheader(curator)

    tomer = curator.make_tome(
        new_tome_name,
        header_tome_name=sub_header_name,
        ds_reading_instructions=[{"channel": "round_end"}],
    )

    for data, _ in tomer.iterate():
        tomer.concat(data["round_end"])

    df = curator.get_dataframe(new_tome_name)
    keyset = curator.get_keyset(new_tome_name)
    assert isinstance(df, pd.DataFrame)
    assert isinstance(keyset, list)


def test_make_tome_with_options(tmp_path):
    new_tome_name_size_limited = "round_end_size_limited"
    new_tome_name_row_limited = "round_end_row_limited"
    curator = create_curator_instance(tmp_path)
    create_header_and_subheader(curator)
    tomer = curator.make_tome(
        new_tome_name_size_limited,
        header_tome_name=sub_header_name,
        ds_reading_instructions=[{"channel": "round_end"}],
        max_page_size_mb=0.00001,
        limit_check_frequency=1,
    )

    for data, _ in tomer.iterate():
        tomer.concat(data["round_end"])

    df = curator.get_dataframe(new_tome_name_size_limited)
    keyset = curator.get_keyset(new_tome_name_size_limited)
    assert isinstance(df, pd.DataFrame)
    assert isinstance(keyset, list)

    tomer = curator.make_tome(
        new_tome_name_row_limited,
        header_tome_name=sub_header_name,
        ds_reading_instructions=[{"channel": "round_end"}],
        max_page_row_count=1,
        limit_check_frequency=1,
    )

    for data, _ in tomer.iterate():
        tomer.concat(data["round_end"])

    df = curator.get_dataframe(new_tome_name_row_limited)
    keyset = curator.get_keyset(new_tome_name_row_limited)
    assert isinstance(df, pd.DataFrame)
    assert isinstance(keyset, list)


def continue_tomer_generator(
    curator,
    behavior_if_complete="pass",
    behavior_if_partial="continue",
):
    tomer = curator.make_tome(
        continued_tome_name,
        header_tome_name=sub_header_name,
        ds_reading_instructions=[{"channel": "round_end"}],
        max_page_row_count=1,
        limit_check_frequency=1,
        behavior_if_complete=behavior_if_complete,
        behavior_if_partial=behavior_if_partial,
    )
    return tomer


def create_partial_tome(curator):
    create_header_and_subheader(curator)
    tomer = continue_tomer_generator(curator)
    for data, _ in tomer.iterate():
        tomer.concat(data["round_end"])
        break


def create_complete_tome(curator):
    create_header_and_subheader(curator)
    tomer = continue_tomer_generator(curator)
    for data, _ in tomer.iterate():
        tomer.concat(data["round_end"])


def test_make_tome_breaking(tmp_path):
    curator = create_curator_instance(tmp_path)
    create_partial_tome(curator)
    keyset = curator.get_keyset(continued_tome_name)
    assert len(keyset) == 1
    assert curator.get_manifest(continued_tome_name)["isComplete"] is False


def test_make_tome_existing_behavior_pass(tmp_path):
    curator = create_curator_instance(tmp_path)
    create_partial_tome(curator)
    tome_id = curator.get_manifest(continued_tome_name)["id"]
    tomer = continue_tomer_generator(
        curator, continued_tome_name, behavior_if_partial="pass"
    )
    for data, _ in tomer.iterate():
        tomer.concat(data["round_end"])
    keyset = curator.get_keyset(continued_tome_name)
    assert len(keyset) == 1
    assert curator.get_manifest(continued_tome_name)["isComplete"] is False
    assert tome_id == curator.get_manifest(continued_tome_name)["id"]


def test_make_tome_existing_behavior_continue(tmp_path):
    curator = create_curator_instance(tmp_path)
    create_partial_tome(curator)
    tome_id = curator.get_manifest(continued_tome_name)["id"]
    tomer = continue_tomer_generator(
        curator, continued_tome_name, behavior_if_partial="continue"
    )
    for data, _ in tomer.iterate():
        tomer.concat(data["round_end"])
        break
    keyset = curator.get_keyset(continued_tome_name)
    assert len(keyset) == 2
    assert tome_id == curator.get_manifest(continued_tome_name)["id"]
    assert curator.get_manifest(continued_tome_name)["isComplete"] is False


def test_make_tome_existing_behavior_overwrite(tmp_path):
    curator = create_curator_instance(tmp_path)
    create_partial_tome(curator)
    tomer = continue_tomer_generator(
        curator, continued_tome_name, behavior_if_partial="overwrite"
    )
    tome_id = curator.get_manifest(continued_tome_name)["id"]
    for data, _ in tomer.iterate():
        tomer.concat(data["round_end"])
    assert tome_id != curator.get_manifest(continued_tome_name)["id"]
    assert curator.get_manifest(continued_tome_name)["isComplete"] is True


def test_make_tome_complete_behavior_continue(tmp_path):
    curator = create_curator_instance(tmp_path)
    create_complete_tome(curator)
    tome_id = curator.get_manifest(continued_tome_name)["id"]
    tomer = continue_tomer_generator(curator, behavior_if_complete="continue")
    for data, _ in tomer.iterate():
        tomer.concat(data["round_end"])
    assert tome_id == curator.get_manifest(continued_tome_name)["id"]


def test_make_tome_complete_behavior_overwrite(tmp_path):
    curator = create_curator_instance(tmp_path)
    create_complete_tome(curator)
    tome_id = curator.get_manifest(continued_tome_name)["id"]
    tomer = continue_tomer_generator(curator, behavior_if_complete="overwrite")
    for data, _ in tomer.iterate():
        tomer.concat(data["round_end"])
    assert tome_id != curator.get_manifest(continued_tome_name)["id"]


def test_get_random_match(tmp_path):
    tmp_path = str(tmp_path)
    curator = create_curator_instance(tmp_path)
    create_header_and_subheader(curator)

    loader = curator.get_random_match()

    manifest = loader.manifest
    data = loader.get_channels()

    assert len(manifest["channels"]) == len(data)



import os
import structlog


def get_page_path_fs(path, subtype, page):
    return os.path.join(path, page[subtype]["key"])


def filter_ds_reader_logs(_, __, event_dict):
    if event_dict.get("client") == "ds_reader_fs":
        raise structlog.DropEvent
    return event_dict


def valid_key_part(part):
    if part is None or part == "":
        return False
    return True


def make_key(my_list):
    return "/".join([arg for arg in my_list if valid_key_part(arg)])



import os
import random
import warnings
from typing import List
import structlog
import pandas as pd

from .header_tome import create_header_tome_from_fs, create_subheader_tome_from_fs

from .loader import TomeLoader
from .scribe import TomeScribe
from .manifest import TomeManifest
from .maker import TomeMaker
from .writer_fs import TomeWriterFs
from .reader_fs import TomeReaderFs
from .header_copier_fs import HeaderTomeCopierFs

from ..ds_io import ChannelInstruction, DsReaderFs, GameDsLoader


class TomeCuratorFs:
    """
    Simple API to manage tomes.

    Parameters
    ----------
    default_header_name : str, default=from env (PURESKILLGG_TOME_DEFAULT_HEADER_NAME)
        Name of the default header.
    ds_type : str, default=from env (PURESKILLGG_TOME_DS_TYPE)
        Type of data science file to read from `ds_collection_root_path`.
    tome_collection_root_path : str, default=from env (PURESKILLGG_TOME_COLLECTION_PATH)
        Path leading to where tomes will be stored.
    ds_collection_root_path : str, default=from env (PURESKILLGG_TOME_DS_COLLECTION_PATH)
        Path leading to a series of (possibly nested) folders containing game Data Science files.
    log : structlog.stdlib.BoundLogger
        Logger used for logging logs.

    See Also
    --------
    TomeMaker : Make tomes.
    """

    def __init__(
        self,
        *,
        default_header_name: str = None,
        ds_type: str = None,
        tome_collection_root_path: str = None,
        ds_collection_root_path: str = None,
        log: structlog.stdlib.BoundLogger = None,
    ):
        self._log = log if log is not None else structlog.get_logger()
        self._default_header_name = get_env_option(
            "default_header_name", default_header_name
        )
        self._ds_type = get_env_option("ds_type", ds_type)
        self._tome_collection_root_path = get_env_option(
            "collection_path", tome_collection_root_path
        )
        self._ds_collection_root_path = get_env_option(
            "ds_collection_path", ds_collection_root_path
        )

    def create_header_tome(
        self, tome_name: str = None, /, *, path_depth=None
    ) -> TomeLoader:
        """
        Create the header tome.

        Parameters
        ----------
        tome_name : str, default=`default_header_name`
            Name of the header that will be created.
        path depth : int, default=4
            DEPRECATED. Please do not use. Search is recursive.

        Returns
        -------
        TomeLoader
            Loader for the header tome just created.
        """
        if path_depth is not None:
            warnings.warn(
                "Warning: the keyword path_depth is deprecated. It is not needed."
            )
        name = tome_name if tome_name is not None else self._default_header_name
        return create_header_tome_from_fs(
            name,
            ds_type=self._ds_type,
            tome_collection_root_path=self._tome_collection_root_path,
            ds_collection_root_path=self._ds_collection_root_path,
            log=self._log,
        )

    def create_subheader_tome(
        self,
        tome_name: str,
        selector: callable = lambda df: [True] * len(df),
        /,
        *,
        src_tome_name: str = None,
    ) -> TomeLoader:
        """
        Create a subheader tome.

        Parameters
        ----------
        tome_name : str, default=`default_header_name`
            Name of the header that will be created.
        selector : callable, default=lambda to select all rows
            The selector is passed directly through to the header
            dataframe and the final subheader tome will be equal to
            `header_dataframe.loc[selector]`.
        src_tome_name : str, default=`default_header_name`
            Source header file. Should be the same as the
            default header name in most cases.

        Returns
        -------
        TomeLoader
            Loader for the header tome just created.
        """
        src_name = (
            src_tome_name if src_tome_name is not None else self._default_header_name
        )
        return create_subheader_tome_from_fs(
            tome_name,
            src_tome_name=src_name,
            selector=selector,
            tome_collection_root_path=self._tome_collection_root_path,
            log=self._log,
        )

    def get_dataframe(self, tome_name: str) -> pd.DataFrame:
        """
        Get the dataframe from a tome.

        Parameters
        ----------
        tome_name : str
            Name of the tome.

        Returns
        -------
        pd.DataFrame
            Pandas dataframe containing the tome's data.
        """
        loader = self.get_loader(tome_name)
        return loader.get_dataframe()

    def get_keyset(self, tome_name: str) -> list:
        """
        Get the keyset from a tome.

        Parameters
        ----------
        tome_name : str
            Name of the tome.

        Returns
        -------
        list
            List containing the tome's keyset.
        """
        loader = self.get_loader(tome_name)
        return loader.get_keyset()

    def get_manifest(self, tome_name: str) -> dict:
        """
        Get the manifest from a tome.

        Parameters
        ----------
        tome_name : str
            Name of the tome.

        Returns
        -------
        dict
            Dictionary containing the tome's manifest.
        """
        loader = self.get_loader(tome_name)
        return loader.manifest

    def iterate_pages(self, tome_name: str) -> TomeLoader.iterate_pages:
        """
        Iterate through pages of a tome.

        Parameters
        ----------
        tome_name : str
            Name of the tome.

        Returns
        -------
        TomeLoader.iterate_pages
            Iterator for pages from TomeLoader.
        """
        loader = self.get_loader(tome_name)
        return loader.iterate_pages()

    def get_loader(self, tome_name: str) -> TomeLoader:
        """
        Get the loader for a tome.

        Parameters
        ----------
        tome_name : str
            Name of the tome.

        Returns
        -------
        TomeLoader
            The TomeLoader instance for this tome.
        """
        reader = TomeReaderFs(
            root_path=self._tome_collection_root_path,
            manifest_key="/".join(["tome", self._ds_type, tome_name, "tome"]),
            log=self._log,
        )
        loader = TomeLoader(reader=reader, log=self._log)
        return loader

    def get_header_loader(self) -> TomeLoader:
        """
        Get the loader for the header tome.

        Returns
        -------
        TomeLoader
            The TomeLoader instance for the header tome.
        """
        return self.get_loader(self._default_header_name)

    def get_random_match(self, subheader_name: str = None, /):
        """
        Get the data for a single random match.

        Parameters
        ----------
        subheader_name : str, default=from env (PURESKILLGG_TOME_DEFAULT_HEADER_NAME)
            Name of the subheader to use. Not specifying this will use the default header.

        Returns
        -------
        GameDsLoader
            The GameDsLoader instance for this tome.
        """
        loader = (
            self.get_header_loader()
            if subheader_name is None
            else self.get_loader(subheader_name)
        )

        keyset = loader.get_keyset()

        index = random.choice(range(0, len(keyset)))
        return self.get_match_by_index(index, subheader_name)

    def get_match_by_index(
        self, index: int = 0, subheader_name: str = None, /
    ) -> GameDsLoader:
        """
        Get the data for a single match.

        Parameters
        ----------
        index : int, default=0
            Index of the match to load.
        subheader_name : str, default=from env (PURESKILLGG_TOME_DEFAULT_HEADER_NAME)
            Name of the subheader to use. Not specifying this will use the default header.

        Returns
        -------
        GameDsLoader
            The GameDsLoader instance for this tome.
        """
        loader = (
            self.get_header_loader()
            if subheader_name is None
            else self.get_loader(subheader_name)
        )

        df_header = loader.get_dataframe()

        csds_reader = DsReaderFs(
            root_path=self._ds_collection_root_path,
            manifest_key=df_header["key"][index],
        )

        csds_loader = GameDsLoader(reader=csds_reader)

        return csds_loader

    # pylint: disable=too-many-locals
    def make_tome(
        self,
        tome_name: str,
        /,
        *,
        header_tome_name: str = None,
        ds_reading_instructions: List[ChannelInstruction] = None,
        max_page_size_mb: float = None,
        max_page_row_count: int = None,
        limit_check_frequency: int = 100,
        **kwargs,
    ) -> TomeMaker:
        """
        Make a tome.

        Parameters
        ----------
        tome_name : str
            Name of the tome.
        ds_reading_instructions : ChannelInstruction, default=None
            Instructions on how to read in each DS file. Default value
            of None will read all channels and columns.
        max_page_size_mb : float, default = None
            Max size *in memory* that a tome can be. Generally it will be
            2-10x smaller on disk. By default the scribe will not check.
        max_page_row_count : int, default = None
            Max number of rows that a tome can be. By default it will
            not have a max row count.
        limit_check_frequency : int
            How often the scribe should check if it exceeded the max size or
            max row count. This should be a multiple of the print frequency
            which is set to 100.
        **kwargs:
            Keywords passed through to the TomeMaker.

        Returns
        -------
        TomeMaker
            The TomeMaker instance to make a tome.
        """
        header_name = (
            header_tome_name
            if header_tome_name is not None
            else self._default_header_name
        )
        name = tome_name

        existing_tome_loader = self.get_loader(name)

        header_loader = self.get_loader(header_name)
        writer = TomeWriterFs(
            root_path=self._tome_collection_root_path,
            log=self._log,
        )
        manifest = TomeManifest(
            tome_name=name,
            ds_type=self._ds_type,
            header_tome_name=header_name,
            log=self._log,
        )
        scribe = TomeScribe(
            manifest=manifest,
            writer=writer,
            max_page_size_mb=max_page_size_mb,
            max_page_row_count=max_page_row_count,
            limit_check_frequency=limit_check_frequency,
            log=self._log,
        )
        header_copier = HeaderTomeCopierFs(
            src_tome_name=header_tome_name,
            tome_collection_root_path=self._tome_collection_root_path,
            dest_tome_name=name,
            ds_type=self._ds_type,
            log=self._log,
        )

        tomer = TomeMaker(
            header_loader=header_loader,
            scribe=scribe,
            ds_reading_instructions=ds_reading_instructions,
            ds_type=self._ds_type,
            ds_collection_root_path=self._ds_collection_root_path,
            tome_loader=existing_tome_loader,
            header_copier=header_copier,
            **kwargs,
            log=self._log,
        )

        return tomer


def get_env_option(name, value):
    if value is not None:
        return value
    env_prefix = "pureskillgg_tome"
    key = "_".join([env_prefix, name]).upper()
    env_value = os.environ.get(key)
    if env_value is not None:
        return env_value
    raise Exception(f"Missing option {name}, pass in or set {key} in environment")



import pickle

from .s3_model import S3Model


class S3Scikit(S3Model):
    def __init__(self, *, model, log):
        super().__init__(model, log)
        self._model_type = model["model_type"]
        self._key = model["key"]
        self._loaded_model = None
        self._log = log.bind(
            client="s3_scikit",
            key=self._key,
            bucket=self._bucket,
            model_name=self.model_name,
        )

    def _load_model(self):
        if self._res_type == "application/x-pickle":
            model = self._read_pickle()
            return model
        raise Exception(f"Unknown res_type {self._res_type}")

    def _read_pickle(self):
        my_pickle = pickle.loads(
            self._s3_client.get_object(Bucket=self._bucket, Key=self._key)[
                "Body"
            ].read()
        )
        return my_pickle

    def _use_model(self, dataframe):
        if self._model_type == "MiniBatchKMeans":
            labels = self._loaded_model.predict(dataframe)
            return labels
        raise Exception(f"Unknown model_type {self._model_type}")

    def invoke(self, dataframe):
        self._log.debug("Invoke: Start")
        if self._loaded_model is None:
            self._loaded_model = self._load_model()
        labels = self._use_model(dataframe)
        return labels



from .api import (
    get_adx_dataset_revisions,
    download_adx_dataset_revision,
    export_single_adx_dataset_revision_to_s3,
    export_multiple_adx_dataset_revisions_to_s3,
    enable_auto_exporting_adx_dataset_revisions_to_s3,
    disable_auto_exporting_adx_dataset_revisions_to_s3,
)



import os
import time
import urllib

import structlog


class AdxDatasetWriterFs:
    def __init__(self, *, root_path, prefix, log=None):
        self._log = log if log is not None else structlog.get_logger()
        self._log = self._log.bind(
            client="adx_dataset_writer_fs", root_path=root_path, prefix=prefix
        )
        self._path = root_path
        self._prefix = prefix

    def export_revision(self, client, dataset_id, revision_id):
        assets = get_assets(client, dataset_id, revision_id)
        if len(assets) > 500:
            self._log.warn(
                f"This revision contains {len(assets)} assets and will take a long time to download. Seriously, you are not gonna want to wait for this. Export to S3 by using export_multiple_adx_dataset_revisions_to_s3 instead."
            )
        for asset in assets:
            self._download_asset(client, asset)

    def auto_export_revisions(self, client, dataset_id):
        raise NotImplementedError("Cannot auto export revisions to file system.")

    def _download_asset(self, client, asset):
        asset_name = asset.get("Name")
        output_path = os.path.join(self._get_output_path(), *asset_name.split("/"))
        output_dir = os.path.dirname(output_path)
        os.makedirs(output_dir, exist_ok=True)

        if os.path.isfile(output_path):
            return

        url = get_download_asset_signed_url(client, asset)
        urllib.request.urlretrieve(url, output_path)
        self._log.info("Downloaded Asset", path=output_path, asset_name=asset_name)

    def _get_output_path(self):
        if self._prefix is None:
            return self._path
        return os.path.join(self._path, self._prefix)


def get_assets(client, dataset_id, revision_id):
    paginator = client.get_paginator("list_revision_assets")
    pages = paginator.paginate(DataSetId=dataset_id, RevisionId=revision_id)
    assets = []
    for page in pages:
        assets.extend(page["Assets"])
    return assets


def get_download_asset_signed_url(client, asset):
    job = client.create_job(
        Type="EXPORT_ASSET_TO_SIGNED_URL",
        Details={
            "ExportAssetToSignedUrl": {
                "AssetId": asset.get("Id"),
                "DataSetId": asset.get("DataSetId"),
                "RevisionId": asset.get("RevisionId"),
            }
        },
    )

    job_id = job.get("Id")
    client.start_job(JobId=job_id)

    while True:
        time.sleep(2)
        job = client.get_job(JobId=job_id)

        if job.get("State") == "ERROR":
            message = job.get("Errors")[0].get("Message")
            raise Exception(f"Job {job_id} failed: {message}")

        if job.get("State") == "CANCELLED":
            raise Exception(f"Job {job_id} was cancelled.")

        if job.get("State") == "TIMED_OUT":
            raise Exception(f"Job {job_id} timed out.")

        if job.get("State") == "COMPLETED":
            return job.get("Details").get("ExportAssetToSignedUrl").get("SignedUrl")



import json

from .s3_model import S3Model


class S3Hashmap(S3Model):
    def __init__(self, *, model, log):
        super().__init__(model, log)
        self._key = model["key"]
        self._log = log.bind(
            client="s3_hashmap",
            key=self._key,
            bucket=self._bucket,
            model_name=self.model_name,
        )

    def _load_model(self):
        if self._res_type == "application/json":
            self._model_data = self._read_json()
        else:
            raise Exception("Unknown res type")

    def _read_json(self):
        res = self._s3_client.get_object(Bucket=self._bucket, Key=self._key)
        body = res["Body"].read().decode("utf-8")
        return json.loads(body)

    def invoke(self, key):
        self._log.debug("Invoke: Start")
        if self._model_data is None:
            self._log.info("Invoke: Load")
            self._load_model()
        return self._model_data.get(key, None)



# pylint: disable=missing-docstring
# pylint: disable=unused-import
# pylint: disable=unused-argument
# pylint: disable=no-self-use
# pylint: disable=protected-access

from glob import glob

import os
import pandas as pd
import pytest
from .game_ds_loader import GameDsLoader
from .reader_fs import DsReaderFs
from .normalize_instructions import normalize_instructions


def create_test_game_ds_loader():
    root = "fixtures"
    manifest_key = "csds/2022/05/15/63cc7181-07c9-42fd-ade4-4eeb2cf4db6f/csds"
    reader = DsReaderFs(root_path=root, manifest_key=manifest_key)
    loader = GameDsLoader(reader=reader)
    return loader


def test_manifest():
    loader = create_test_game_ds_loader()
    assert "channels" in loader.manifest
    assert isinstance(loader.metadata, dict)


def test_metadata():
    loader = create_test_game_ds_loader()
    assert isinstance(loader.metadata, dict)


def test_blank_instructions():
    loader = create_test_game_ds_loader()
    data = loader.get_channels()
    assert len(data.keys()) == len(loader.manifest["channels"])


def test_reading_channels_with_instructions():
    loader = create_test_game_ds_loader()
    data = loader.get_channels([{"channel": "round_end"}])
    assert "round_end" in data
    assert isinstance(data["round_end"], pd.DataFrame)


def test_single_channel():
    loader = create_test_game_ds_loader()
    df = loader.get_channel({"channel": "round_end"})
    assert isinstance(df, pd.DataFrame)


def test_channels_with_column():
    loader = create_test_game_ds_loader()
    data = loader.get_channels([{"channel": "round_end", "columns": ["tick"]}])
    assert len(data["round_end"].columns) == 1


def test_single_channel_with_column():
    loader = create_test_game_ds_loader()
    df = loader.get_channel({"channel": "round_end", "columns": ["tick"]})
    assert isinstance(df, pd.DataFrame)
    assert len(df.columns) == 1



import json


def find_matching_model(models, filter_dict):
    output = []
    for model in models:
        key = model.pop("key")
        if model == filter_dict:
            output.append(model)
        model["key"] = key  # MUST restore the key

    if len(output) > 1:
        raise Exception(
            f"Found {len(models)} matches in model set but expected only 1 to match filter {json.dumps(filter_dict)}"
        )
    if len(output) == 0:
        return None
    return output[0]



import boto3


class S3Model:
    def __init__(self, model, log):
        self.model_name = model["model_name"]
        self.model_parameters = model.get("parameters", {})
        self._bucket = model["bucket"]
        self._res_type = model["res_type"]
        self._model_data = None
        self._s3_client = boto3.client("s3")
        self._prefix = ""
        self._extension = ""
        self._log = log

    def _get_key(self, key):
        return "/".join([self._prefix, ".".join([key, self._extension])])



from typing import TypedDict, List, Dict, Optional, Any, Callable, Union
from collections.abc import Iterable
import structlog

import pandas as pd

from .reader_fs import DsReaderFs
from .reader_s3 import DsReaderS3
from .normalize_instructions import normalize_instructions


class ChannelInstruction(TypedDict):
    channel: str
    columns: Optional[List[str]]


class GameDsLoader:
    def __init__(self, *, reader: Union[DsReaderFs, DsReaderS3], log: object = None):
        self._reader = reader
        self._log = log if log is not None else structlog.get_logger()
        self._metadata = None
        self._manifest = None

    @property
    def manifest(self):
        """Root channel data"""
        self._load()
        return self._manifest

    @property
    def metadata(self):
        """Metadata"""
        self._load()
        return self._metadata

    def _load(self) -> None:
        is_loaded = self._manifest is not None and self._metadata is not None
        if is_loaded:
            return
        self._manifest = self._reader.read_manifest()
        self._metadata = self._reader.read_metadata()

    def get_channels(
        self,
        channel_instructions: Optional[List[ChannelInstruction]] = None,
    ) -> Dict[str, pd.DataFrame]:
        """Read in data from a ds object

        Args:
            channel_instructions (dict, optional): Instructions on how to read the ds
                file where multiple instructions for a channel will be smartly combined.
                Instructions indicate how to read each channel.
                Each element of a set of instructions must contain a key
                called `channel` and optionally have a key named `columns`
                which is a List of columns to be read in. Not including the
                `columns` key implies you want to read in all columns.
                Omitting this keyword reads all channels and columns


        Returns:
            Data (dict): dictionary where each key is the channel name and the value
                is a Pandas data frame.


        Examples:
            loader = GameDsLoader(reader=reader, log=log)

            # read all channels
            data = loader.get_channels()

            # read one specific channel
            data = loader.get_channels([{"channel":"ch1"}])

            # read two channels
            data = loader.get_channels([
                {"channel":"ch1"},{"channel":"ch2"}
            ])

            # read channels with specific columns
            data = loader.get_channels([
                {"channel":"ch1","columns":["col1","col3"]}
            ])

            # combine sets of instructions
            data = loader.get_channels([
                {"channel":"ch1","columns":["col1","col3"]},
                {"channel":"ch1","columns":["col1","col2"]}
            ])
            # same as above
            data = loader.get_channels([
                {"channel":"ch1","columns":["col1","col2","col3"]}
            ])
        """
        self._load()

        instructions = self._normalize_channel_instructions(channel_instructions)

        output = {}
        for instruction in instructions:
            df = self.get_channel(instruction)
            output[instruction["channel"]] = df
        return output

    def get_channel(self, instruction: ChannelInstruction) -> (str, pd.DataFrame):
        """Read in one channel from a ds object

        Args:
            instruction (dict): dictionary containing a key called
                `channel` with the value is the name of the channel
                to be read in. optionally have a key named `columns`
                which is a List of columns to be read in.


        Returns:
            Pandas data frame containing the data you requested and the channel name.

        """
        self._load()

        channel = self._find_channel_by_name(instruction["channel"])

        content_type = channel["contentType"]
        if content_type != "application/x-parquet":
            raise Exception(f"Unsupported content type {content_type}")

        df = self._reader.read_parquet_channel(
            channel, columns=instruction.get("columns")
        )
        return df

    def _find_channel_by_name(self, channel_name: str, /) -> Dict:
        channel = first_true(
            self._manifest["channels"],
            None,
            lambda channel: channel["channel"] == channel_name,
        )
        if channel is None:
            raise Exception(f"Channel {channel_name} not found in replay.")
        return channel

    def _normalize_channel_instructions(
        self, channel_instructions: Union[List[ChannelInstruction], None]
    ) -> List[ChannelInstruction]:
        if not is_valid_instructions(channel_instructions):
            raise Exception(
                f"The instructions were not a list of instructions or None: {channel_instructions}"
            )
        if isinstance(channel_instructions, list):
            return normalize_instructions(channel_instructions)
        instructions = []
        for channel in self._manifest["channels"]:
            single_instruction: ChannelInstruction = {"channel": channel["channel"]}
            instructions.append(single_instruction)
        return instructions


def is_valid_instructions(channel_instructions: Union[List[ChannelInstruction], None]):
    if isinstance(channel_instructions, list):
        return True
    if channel_instructions is None:
        return True
    return False


def first_true(
    iterable: Iterable, /, default: Any = False, pred: Optional[Callable] = None
) -> Any:
    return next(filter(pred, iterable), default)



import os

from pathlib import Path
import structlog
import rapidjson
import pandas as pd

from .constants import get_page_path_fs


class TomeWriterFs:
    def __init__(
        self,
        *,
        root_path,
        prefix=None,
        log=None,
    ):
        self._log = log if log is not None else structlog.get_logger()
        self._log = self._log.bind(
            client="tome_writer_fs",
            root_path=root_path,
            prefix=prefix,
        )
        self._root_path = root_path
        self._prefix = prefix
        self._parquet_compression = "gzip"

    def write_manifest(self, manifest):
        file_location = os.path.join(
            self._root_path, add_prefix(manifest["key"], self._prefix)
        )
        ensure_dir(file_location)
        self._log.info("Write Manifest: Start")

        self._write_json(file_location, manifest)

    def write_page(self, page, dataframe, keyset):
        ensure_dir(self._get_page_key("dataframe", page))
        self._log.info("Write Page Start", page_number=page["number"])
        self._write_dataframe(page, dataframe)
        self._write_keyset(page, keyset)

    def _write_dataframe(self, page, dataframe):
        key = self._get_page_key("dataframe", page)

        content_type = page["dataframe"]["contentType"]
        if content_type != "application/x-parquet":
            raise Exception(f"Unsupported content type {content_type}")

        self._log.info("Write Dataframe: Start", page_number=page["number"])
        self._write_parquet(key, dataframe)

    def _write_keyset(self, page, keyset):
        key = self._get_page_key("keyset", page)

        content_type = page["keyset"]["contentType"]
        if content_type != "application/x-parquet":
            raise Exception(f"Unsupported content type {content_type}")

        self._log.info("Write keyset: Start", page_number=page["number"])
        df = pd.DataFrame(
            keyset, columns=["_"]
        )  # must have string column name for parquet
        self._write_parquet(key, df)

    def _get_page_key(self, subtype, page):
        return get_page_path_fs(self._root_path, subtype, page)

    def _write_parquet(self, key: str, df: pd.DataFrame) -> None:
        self._log.debug("Write parquet", key=key)
        df.to_parquet(key, compression=self._parquet_compression)

    # pylint: disable=no-self-use
    def _write_json(self, key, data):
        with open(key, "w", encoding="utf-8") as file:
            rapidjson.dump(data, file)


def add_prefix(key, prefix, /) -> str:
    if prefix is None:
        return key
    return os.path.join(*[*prefix.split("/"), key])


def ensure_dir(file_location):
    folder = Path(file_location).parent
    if not os.path.isdir(folder):
        os.makedirs(folder)



import os
from glob import glob
import structlog

from ..ds_io import DsReaderFs, GameDsLoader
from .loader import TomeLoader
from .scribe import TomeScribe
from .manifest import TomeManifest
from .writer_fs import TomeWriterFs
from .reader_fs import TomeReaderFs
from .constants import filter_ds_reader_logs


def default_tome_name():
    return "header"


# pylint: disable=unused-argument
def create_header_tome_from_fs(
    tome_name=None,
    /,
    *,
    ds_type="csds",
    tome_collection_root_path="tomes",
    ds_collection_root_path="data",
    path_depth=None,
    log=None,
):
    """Make the header tome"""
    name = default_tome_name() if tome_name is None else tome_name
    log = (
        log
        if log is not None
        else structlog.wrap_logger(
            structlog.get_logger(), processors=[filter_ds_reader_logs]
        )
    )

    writer = TomeWriterFs(root_path=tome_collection_root_path, log=log)
    tome_manifest = TomeManifest(tome_name=name, ds_type=ds_type, is_header=True)
    scribe = TomeScribe(manifest=tome_manifest, writer=writer, log=log)

    scribe.start()

    for manifest_key_path in get_manifest_key_paths_from_glob(
        ds_collection_root_path, ds_type
    ):
        ds_loader = fetch_ds_loader_from_fs(
            ds_collection_root_path, manifest_key_path, log
        )
        df = ds_loader.get_channel({"channel": "header"})
        df["key"] = ds_loader.manifest["key"]
        df["match_id"] = ds_loader.manifest["id"]
        scribe.concat(df, ds_loader.manifest["key"])

    scribe.finish()

    reader = TomeReaderFs(
        root_path=tome_collection_root_path,
        manifest_key="/".join(["tome", ds_type, tome_name, "tome"]),
        log=log,
    )

    return TomeLoader(reader=reader, log=log)


# pylint: disable=too-many-locals
def create_subheader_tome_from_fs(
    name,
    /,
    src_tome_name=None,
    selector=lambda df: [True] * len(df),
    *,
    tome_collection_root_path="tomes",
    ds_type="csds",
    is_copied_header=False,
    preserve_src_id=False,
    log=None,
):
    """
    selector is passed to filter out rows from the header
    by doing df = df.loc[selector]
    """
    log = log if log is not None else structlog.get_logger()
    src_name = default_tome_name() if src_tome_name is None else src_tome_name

    src_reader = TomeReaderFs(
        root_path=tome_collection_root_path,
        manifest_key="/".join(["tome", ds_type, src_name, "tome"]),
        log=log,
    )
    src_loader = TomeLoader(reader=src_reader, log=log)

    writer = TomeWriterFs(
        root_path=tome_collection_root_path,
        log=log,
    )

    manifest = TomeManifest(
        tome_name=name,
        ds_type=src_loader.manifest["dsType"],
        header_tome_name=src_name,
        is_header=True,
        src_id=src_loader.manifest["id"] if preserve_src_id else None,
        is_copied_header=is_copied_header,
    )
    scribe = TomeScribe(writer=writer, manifest=manifest, log=log)
    scribe.start()
    df = src_loader.get_dataframe().loc[selector]
    scribe.concat(df, list(df["key"]))
    scribe.finish()

    reader = TomeReaderFs(
        root_path=tome_collection_root_path,
        manifest_key=manifest.get()["key"],
        log=log,
    )

    return TomeLoader(reader=reader, log=log)


def get_manifest_key_paths_from_glob(ds_root_path, ds_type):
    root_path = os.path.normpath(ds_root_path) + os.sep
    paths = glob(os.path.join(root_path, "*", "**", ds_type), recursive=True)
    paths = [os.path.normpath(path) for path in paths]
    manifest_keys = [
        path[len(os.path.commonprefix([root_path, path])) :] for path in paths
    ]
    manifest_keys = set(manifest_keys)
    return manifest_keys


def fetch_ds_loader_from_fs(root_path, manifest_key, log):
    ds_reader = DsReaderFs(
        root_path=root_path,
        manifest_key=manifest_key,
        log=log,
    )
    ds_loader = GameDsLoader(reader=ds_reader, log=log)
    return ds_loader



from typing import Dict
import pandas as pd


def handle_value_error(value_error: object, channel: Dict) -> pd.DataFrame:
    if str(value_error) == "need at least one array to concatenate":
        return pd.DataFrame({col["name"]: [] for col in channel["columns"]})
    raise Exception(f"Couldn't read in parquet file {channel['channel']}")



import dateutil

import structlog
import boto3


class AdxDataset:
    def __init__(self, *, dataset_id, writer, log=None):
        self._log = log if log is not None else structlog.get_logger()
        self._log = self._log.bind(
            client="adx_dataset",
            dataset_id=dataset_id,
        )
        self._client = boto3.client("dataexchange")
        self.dataset_id = dataset_id
        self.dataset_name = None
        self._writer = writer
        self._dataset = None

    def get_latest_revision(self):
        self._init()
        res = self._client.list_data_set_revisions(DataSetId=self.dataset_id)
        revisions = res.get("Revisions", [])
        if len(revisions) == 0:
            return None
        return revisions[0]

    def get_revisions(self, start_date=None, end_date=None, /):
        self._init()
        paginator = self._client.get_paginator("list_data_set_revisions")
        pages = paginator.paginate(DataSetId=self.dataset_id)
        revisions = []
        for page in pages:
            revisions.extend(page["Revisions"])
        return [
            rev
            for rev in revisions
            if is_date_between(rev["Comment"], start_date, end_date)
        ]

    def export_revisions(self, start_date=None, end_date=None, /):
        revisions = self.get_revisions(start_date, end_date)
        self._log.info(
            "Selected Revisions",
            start_date=start_date,
            end_date=end_date,
            count=len(revisions),
        )
        for revision in revisions:
            self.export_revision(revision["Id"])

    def export_revision(self, revision_id):
        self._init()
        revision = self._client.get_revision(
            DataSetId=self.dataset_id, RevisionId=revision_id
        )
        log = self._log.bind(revision_id=revision_id, comment=revision.get("Comment"))
        try:
            log.info("Export Revision: Start")
            self._writer.export_revision(self._client, self.dataset_id, revision_id)
            log.info("Export Revision: Success")
        except BaseException as err:
            log.error("Export Revision: Fail", exec_info=err)
            raise

    def auto_export_revisions(self):
        self._init()
        return self._writer.auto_export_revisions(self._client, self.dataset_id)

    def delete_all_auto_revision_exports(self):
        self._init()
        paginator = self._client.get_paginator("list_event_actions")
        pages = paginator.paginate()
        actions = []
        for page in pages:
            actions.extend(page["EventActions"])

        for action in actions:
            dataset_id = (
                action.get("Event", {}).get("RevisionPublished", {}).get("DataSetId")
            )
            if dataset_id == self.dataset_id:
                self._client.delete_event_action(EventActionId=action["Id"])
                self._log.info("Deleted Action", event_action_id=action["Id"])

    def _init(self):
        if self._dataset is not None:
            return

        res = self._client.get_data_set(DataSetId=self.dataset_id)
        self.dataset_name = res.get("Name")
        self._log = self._log.bind(dataset_name=self.dataset_name)


def is_date_between(date, start_date, end_date):
    start = dateutil.parser.isoparse(start_date if start_date is not None else "1900")
    end = dateutil.parser.isoparse(end_date if end_date is not None else "4000")
    try:
        middle = dateutil.parser.isoparse(date.split("T")[0])
        return start <= middle < end
    except dateutil.parser.ParserError:
        return False



"""
Python Data Science Development Kit.
"""
from .ds_models import create_ds_models
from .ds_io.reader_s3 import DsReaderS3
from .ds_io.reader_fs import DsReaderFs
from .ds_io.game_ds_loader import GameDsLoader
from .tome import TomeCuratorFs, create_tome_curator
from .adx import (
    get_adx_dataset_revisions,
    download_adx_dataset_revision,
    export_single_adx_dataset_revision_to_s3,
    export_multiple_adx_dataset_revisions_to_s3,
    enable_auto_exporting_adx_dataset_revisions_to_s3,
    disable_auto_exporting_adx_dataset_revisions_to_s3,
)



from datetime import datetime, timezone
from uuid import uuid4
import structlog
from .constants import make_key

# pylint: disable=too-many-arguments
class TomeManifest:
    def __init__(
        self,
        *,
        tome_name,
        ds_type,
        is_header=False,
        header_tome_name=None,
        src_id=None,
        is_copied_header=False,
        log=None,
    ):
        self._log = log if log is not None else structlog.get_logger()
        self._tome_name = tome_name
        self._ds_type = ds_type
        self._is_copied_header = is_copied_header
        self._data = create_manifest(
            self._tome_name,
            self._ds_type,
            is_header,
            header_tome_name,
            src_id,
            is_copied_header=self._is_copied_header,
        )
        self._current_page_start_time = None
        self._current_page_end_time = None

    def set(self, data):
        self._data = data

    def start_page(self):
        self._current_page_start_time = now()

    def end_page(self, page_number):
        self._current_page_end_time = now()
        page = {
            "number": page_number,
            "keyset": {
                "keyset": content_name(page_number, "keyset"),
                "key": make_key(
                    [
                        "tome",
                        self._ds_type,
                        self._tome_name,
                        "header" if self._is_copied_header else "",
                        content_name(page_number, "keyset"),
                    ]
                ),
                "contentType": "application/x-parquet",
            },
            "dataframe": {
                "dataframe": content_name(page_number, "dataframe"),
                "key": make_key(
                    [
                        "tome",
                        self._ds_type,
                        self._tome_name,
                        "header" if self._is_copied_header else "",
                        content_name(page_number, "dataframe"),
                    ]
                ),
                "contentType": "application/x-parquet",
            },
            "createdAt": now_to_iso(),
            "timings": self._calculate_end_page_timings(),
        }
        self._data["pages"].append(page)
        return page

    def finish(self):
        self._data["isComplete"] = True
        self._data["timings"] = self._sum_page_timings()
        self._data["completedAt"] = now_to_iso()

    def get(self):
        return self._data

    def _sum_page_timings(self):
        timings = {
            "seconds": 0,
            "minutes": 0,
            "hours": 0,
            "days": 0,
        }
        for page in self._data["pages"]:
            for key in timings:
                timings[key] += page["timings"][key]
        return timings

    def _calculate_end_page_timings(self):
        time_elapsed = (
            self._current_page_end_time.timestamp()
            - self._current_page_start_time.timestamp()
        )
        return {
            "seconds": time_elapsed,
            "minutes": time_elapsed / 60,
            "hours": time_elapsed / 60 / 60,
            "days": time_elapsed / 60 / 60 / 24,
        }


def content_name(page_number, subtype):
    return f"{subtype}_{str(page_number).zfill(5)}"


def create_manifest(
    tome_name, ds_type, is_header, header_tome_name, src_id, is_copied_header=False
):
    key = make_key(
        ["tome", ds_type, tome_name, "header" if is_copied_header else "", "tome"]
    )
    return {
        "id": str(uuid4()),
        "key": key,
        "tome": tome_name,
        "sourceId": src_id,
        "type": "tome",
        "isHeader": is_header,
        "headerTomeName": header_tome_name,
        "dsType": ds_type,
        "createdAt": now_to_iso(),
        "isComplete": False,
        "pages": [],
    }


def now():
    return datetime.now(timezone.utc)


def now_to_iso():
    return now().isoformat()



import json

import boto3
import pandas as pd


class SagemakerEndpoint:
    def __init__(self, *, model, log):
        self.model_name = model["model_name"]
        self.model_parameters = model.get("parameters", {})
        self.columns = self.model_parameters.get("columns", [])
        self._endpoint_name = model["endpoint_name"]
        self._content_type = model["req_type"]
        self._res_type = model["res_type"]
        self._res_key = model.get("res_key", "data")
        self._log = log.bind(
            client="sagemaker_endpoint",
            endpoint_name=self._endpoint_name,
            model_name=self.model_name,
        )

    def invoke(self, dataframe):
        self._log.info("Invoke: Start")
        runtime = boto3.client("runtime.sagemaker")
        final_data = self._format_data(dataframe)
        response = runtime.invoke_endpoint(
            EndpointName=self._endpoint_name,
            ContentType=self._content_type,
            Body=final_data,
        )

        return self._transform_sagemaker_output(response)

    def _format_data(self, dataframe):
        if self._content_type == "text/csv":
            final_data = dataframe[self.columns].to_csv(index=False, header=False)
        else:
            raise Exception(f"Unknown content type {self._content_type}")
        return final_data

    def _transform_sagemaker_output(self, response):
        if self._res_type == "application/json":
            results = json.loads(response["Body"].read().decode())
            return pd.DataFrame.from_dict(results.get(self._res_key))

        raise Exception(f"Unknown res type {self._res_type}")



from io import BytesIO
from gzip import GzipFile
from typing import List, Dict, Optional, Union

import structlog
import rapidjson
import boto3
import pandas as pd
from .handle_value_error import handle_value_error


class DsReaderS3:
    def __init__(
        self,
        *,
        bucket: str,
        manifest_key: str,
        prefix: Optional[str] = None,
        log: object = None,
    ):
        self._log = log if log is not None else structlog.get_logger()
        self._log = self._log.bind(client="ds_reader_s3", bucket=bucket, prefix=prefix)
        self._bucket = bucket
        self._prefix = prefix
        self._s3_client = boto3.client("s3")
        self._manifest_key = manifest_key

    def read_manifest(self) -> Dict:
        self._log.info("Read Manifest: Start")

        if self._manifest_key is None:
            raise Exception("No manifest key given")

        key = add_prefix(self._manifest_key, self._prefix)
        return self._get_json_object(key)

    def read_metadata(self) -> Dict:
        self._log.info("Read Metadata: Start")

        if self._manifest_key is None:
            raise Exception("No manifest key given")

        key = add_prefix(self._manifest_key, self._prefix)
        return self._get_metadata(key)

    def read_parquet_channel(self, channel: Dict, columns: List[str]) -> pd.DataFrame:
        self._log.info(
            "Read parquet Channel: Start", channel=channel["channel"], columns=columns
        )

        try:
            # UPSTREAM: Flaky, will sometimes throw FileNotFoundError on s3 objects
            # https://github.com/apache/arrow/issues/2192#issuecomment-569813829
            try:
                df = pd.read_parquet(
                    self._get_channel_location(channel), columns=columns
                )
            except FileNotFoundError:
                key = self._get_channel_key(channel)
                obj = self._s3_client.get_object(Bucket=self._bucket, Key=key)
                df = pd.read_parquet(BytesIO(obj["Body"].read()), columns=columns)
        except ValueError as value_error:
            df = handle_value_error(value_error, channel)

        return df

    def _get_json_object(self, key: str, /) -> Dict:
        res = self._s3_client.get_object(Bucket=self._bucket, Key=key)

        content_encoding = res["ContentEncoding"]

        body = res["Body"].read()
        if content_encoding == "gzip":
            body = BytesIO(body)
            body = GzipFile(fileobj=body, mode="rb").read().decode("utf-8")
        else:
            body = body.decode("utf-8")

        return rapidjson.loads(body)

    def _get_metadata(self, key: str, /) -> Dict:
        res = self._s3_client.head_object(Bucket=self._bucket, Key=key)
        return {
            **res["Metadata"],
            "key": key,
            "bucket": self._bucket,
            "content_type": res["ContentType"],
            "last_modified": res["LastModified"].isoformat(),
        }

    def _get_compression(self, channel: Dict, /) -> Union[str, None]:
        key = self._get_channel_key(channel)
        res = self._s3_client.head_object(Bucket=self._bucket, Key=key)

        content_encoding = res["ContentEncoding"]

        if content_encoding == "gzip":
            return "gzip"

        return None

    def _get_channel_location(self, channel: Dict, /) -> str:
        key = self._get_channel_key(channel)
        return f"s3://{self._bucket}/{key}"

    def _get_channel_key(self, channel: Dict) -> str:
        key = channel["key"]
        prefix = self._prefix
        return add_prefix(key, prefix)


def add_prefix(key, prefix: Union[str, None], /) -> str:
    if prefix is None:
        return key
    return "/".join([prefix, key])



from .dataset import AdxDataset
from .writer_fs import AdxDatasetWriterFs
from .writer_s3 import AdxDatasetWriterS3


def get_adx_dataset_revisions(dataset_id, /, *, start_date=None, end_date=None):
    client = AdxDataset(dataset_id=dataset_id, writer=None)
    return client.get_revisions(start_date, end_date)


def download_adx_dataset_revision(
    root_path, dataset_id, revision_id=None, /, *, prefix=None
):
    writer = AdxDatasetWriterFs(root_path=root_path, prefix=prefix)
    client = AdxDataset(dataset_id=dataset_id, writer=writer)

    if revision_id is None:
        rev = client.get_latest_revision()
        if rev is None:
            raise RuntimeError("No revisions in dataset")
        revision_id = rev["Id"]

    client.export_revision(revision_id)


def export_single_adx_dataset_revision_to_s3(
    bucket, dataset_id, revision_id=None, /, *, prefix=None
):
    writer = AdxDatasetWriterS3(bucket=bucket, prefix=prefix)
    client = AdxDataset(dataset_id=dataset_id, writer=writer)

    if revision_id is None:
        rev = client.get_latest_revision()
        if rev is None:
            raise RuntimeError("No revisions in dataset")
        revision_id = rev["Id"]

    client.export_revision(revision_id)


def export_multiple_adx_dataset_revisions_to_s3(
    bucket, dataset_id, /, *, prefix=None, start_date=None, end_date=None
):
    writer = AdxDatasetWriterS3(bucket=bucket, prefix=prefix)
    client = AdxDataset(dataset_id=dataset_id, writer=writer)
    client.export_revisions(start_date, end_date)


def enable_auto_exporting_adx_dataset_revisions_to_s3(
    bucket, dataset_id, /, *, prefix=None
):
    writer = AdxDatasetWriterS3(bucket=bucket, prefix=prefix)
    client = AdxDataset(dataset_id=dataset_id, writer=writer)
    client.auto_export_revisions()


def disable_auto_exporting_adx_dataset_revisions_to_s3(dataset_id):
    client = AdxDataset(dataset_id=dataset_id, writer=None)
    client.delete_all_auto_revision_exports()



import structlog
import pandas as pd


class TomeLoader:
    def __init__(self, *, reader, has_header=True, log: object = None):
        self._reader = reader
        self._log = log if log is not None else structlog.get_logger()
        self._manifest = None
        self._metadata = None
        self._exists = None
        self.has_header = has_header
        self.header = None

        if self.has_header:
            self.header = TomeLoader(
                reader=self._reader.header, has_header=False, log=self._log
            )

    @property
    def metadata(self):
        """Metadata"""
        self._load()
        return self._metadata

    @property
    def exists(self):
        """If the tome exists"""
        return self._reader.exists

    @property
    def manifest(self):
        """Tome manifest"""
        self._load()
        return self._manifest

    def _load(self) -> None:
        is_loaded = self._manifest is not None and self._metadata is not None
        if is_loaded:
            return
        self._metadata = self._reader.read_metadata()
        self._manifest = self._reader.read_manifest()

    def get_dataframe(self):
        self._load()

        dfs = [
            self._reader.read_page_dataframe(page) for page in self.manifest["pages"]
        ]
        df = pd.concat(dfs)
        return df

    def get_keyset(self):
        self._load()

        keyset = []
        for page in self.manifest["pages"]:
            keyset += self._reader.read_page_keyset(page)
        return keyset

    def iterate_pages(self):
        self._load()
        for page in self.manifest["pages"]:
            yield (
                self._reader.read_page_dataframe(page),
                self._reader.read_page_keyset(page),
            )



import os
from pathlib import Path
import pandas as pd
import structlog
import rapidjson
from .constants import (
    get_page_path_fs,
)


class TomeReaderFs:
    def __init__(
        self,
        *,
        root_path,
        prefix=None,
        manifest_key,
        has_header=True,
        log=None,
    ):
        self._log = log if log is not None else structlog.get_logger()
        self._log = self._log.bind(
            client="tome_reader_fs",
            root_path=root_path,
            prefix=prefix,
            manifest_key=manifest_key,
        )

        self._root_path = root_path
        self._prefix = prefix
        self._manifest_key = manifest_key
        self.has_header = has_header
        self.header = None
        if self.has_header:
            self.header = TomeReaderFs(
                root_path=self._root_path,
                manifest_key="/".join(
                    [str(Path(self._manifest_key).parent), "header", "tome"]
                ),
                has_header=False,
                log=self._log,
            )

    @property
    def exists(self):
        """If the tome exists"""
        try:
            self.read_manifest()
        except FileNotFoundError:
            return False
        except:
            self._log.error("There was an error while loading the loader")
            raise
        return True

    def read_manifest(self):
        self._log.info("Read Manifest: Start")
        file_location = os.path.join(
            self._root_path, add_prefix(self._manifest_key, self._prefix)
        )

        with open(file_location, "r", encoding="utf-8") as file:
            data = rapidjson.loads(file.read())
        return data

    # pylint: disable=no-self-use
    def read_metadata(self):
        return {}

    def read_page(self, page):
        dataframe = self.read_page_dataframe(page)
        keyset = self.read_page_keyset(page)
        return dataframe, keyset

    def read_page_keyset(self, page):
        key = self._get_page_key("keyset", page)

        content_type = page["keyset"]["contentType"]
        if content_type != "application/x-parquet":
            raise Exception(f"Unknown content type {content_type}")

        self._log.info("Read keyset: Start", page_number=page["number"])
        df = pd.read_parquet(key)
        return list(df.iloc[:, 0])

    def read_page_dataframe(self, page):
        key = self._get_page_key("dataframe", page)

        content_type = page["dataframe"]["contentType"]
        if content_type != "application/x-parquet":
            raise Exception(f"Unsupported content type {content_type}")

        self._log.info("Read Dataframe: Start", page_number=page["number"])
        df = pd.read_parquet(key)
        return df

    def _get_page_key(self, subtype, page):
        return get_page_path_fs(self._root_path, subtype, page)


def add_prefix(key, prefix, /) -> str:
    if prefix is None:
        return key
    return os.path.join(*[*prefix.split("/"), key])



import time
import structlog
import pandas as pd

from ..ds_io import DsReaderFs, GameDsLoader
from .constants import filter_ds_reader_logs


class TomeMaker:
    def __init__(
        self,
        *,
        header_loader,
        scribe,
        ds_reading_instructions,
        ds_type,
        tome_loader,
        header_copier,
        ds_collection_root_path,
        behavior_if_complete="pass",
        behavior_if_partial="continue",
        print_status_frequency=100,
        reader_class=DsReaderFs,
        log: object = None,
    ):
        self.is_started = False
        self.is_finished = False
        self.keyset = None
        self._header_dataframe = None
        self._log = log if log is not None else structlog.get_logger()
        self._scribe = scribe
        self._ds_type = ds_type
        self._ds_reading_instructions = ds_reading_instructions
        self._header_loader = header_loader
        self._existing_tome_loader = tome_loader
        self._header_copier = header_copier
        self._if_complete = behavior_if_complete
        self._if_partial = behavior_if_partial
        self._print_status_frequency = print_status_frequency
        self._reader_class = reader_class
        self._current_key = None
        self._loaded = False
        self._ds_collection_root_path = ds_collection_root_path

    def iterate(self):
        """Get data for the next key"""
        self._load()
        if len(self.keyset) == 0:
            return
        self._scribe.start()
        self._log = structlog.wrap_logger(self._log, processors=[filter_ds_reader_logs])

        start_time = time.time()
        key_counter = 0
        for key in self.keyset:
            self._current_key = key
            data = self._get_ds_channels_from_fs(key)
            yield data, key
            self._log_status(key_counter, start_time)
            key_counter += 1

        self._finish()

    def concat(self, df: pd.DataFrame):
        """Append a dataframe to tome dataset"""
        self._scribe.concat(df, self._current_key)

    def _finish(self):
        self._scribe.finish()

    def _log_status(self, key_counter: int, start_time: float):
        if key_counter % self._print_status_frequency == 0 and key_counter > 0:
            n_done = key_counter
            n_total = len(self.keyset)
            fraction_complete = n_done / n_total
            seconds_elapsed = time.time() - start_time
            time_remaining = (seconds_elapsed / fraction_complete) - seconds_elapsed
            meta = {
                "page_count": self._scribe.page_counter,
                "page_row_count": self._scribe.page_row_count,
                "page_size_mb": self._scribe.page_size_mb,
                "percent_complete": round(100 * fraction_complete, 3),
                "keys_complete": n_done,
                "minutes_elapsed": int(seconds_elapsed / 60),
                "minutes_remaining": int(time_remaining / 60),
                "hours_elapsed": int(seconds_elapsed / 60 / 60),
                "hours_remaining": int(time_remaining / 60 / 60),
            }
            self._log.info("Tome Maker Update:", **meta)

    def _get_ds_channels_from_fs(self, key):
        root_path = self._ds_collection_root_path
        manifest_key = key
        ds_reader = self._reader_class(
            root_path=root_path,
            manifest_key=manifest_key,
            log=self._log,
        )
        ds_loader = GameDsLoader(reader=ds_reader, log=self._log)
        data = ds_loader.get_channels(self._ds_reading_instructions)
        return data

    def _load_actions(self):
        def continue_tome():
            self._header_dataframe = self._existing_tome_loader.header.get_dataframe()
            existing_keyset = self._existing_tome_loader.get_keyset()
            target_keyset = self._existing_tome_loader.header.get_keyset()
            self.keyset = list(set(target_keyset) - set(existing_keyset))
            self._scribe.set_manifest_data(self._existing_tome_loader.manifest)

        def overwrite():
            self._header_dataframe = self._header_loader.get_dataframe()
            self.keyset = self._header_loader.get_keyset()
            self._copy_header()

        def start_new():
            overwrite()

        def passthrough():
            self.keyset = []

        def fail():
            raise Exception(f"Tome already exists {self._scribe.tome_name}")

        return continue_tome, overwrite, start_new, passthrough, fail

    def _load(self):
        if self._loaded:
            return

        self._header_dataframe = self._header_loader.get_dataframe()

        continue_tome, overwrite, start_new, passthrough, fail = self._load_actions()

        exists = self._existing_tome_loader.exists
        if not exists:
            start_new()
        else:
            is_complete = self._existing_tome_loader.manifest["isComplete"]
            action = {
                "continue": continue_tome,
                "overwrite": overwrite,
                "pass": passthrough,
                "fail": fail,
            }[self._if_complete if is_complete else self._if_partial]
            if is_complete and self._if_complete == "continue":
                action = passthrough
            action()

        self._loaded = True

    def _copy_header(self):
        self._header_copier.copy()



from io import BytesIO

import pandas as pd

from .find import find_matching_model
from .s3_model import S3Model


class S3DataframeSet(S3Model):
    def __init__(self, *, model, log):
        super().__init__(model, log)
        self._prefix = model["prefix"]
        self._extension = model["extension"]
        self._dataframes = model["dataframes"]
        self._model_selected = False
        self._selected_key = None
        self._log = log.bind(
            client="s3_dataframe_set",
            prefix=self._prefix,
            bucket=self._bucket,
            model_name=self.model_name,
        )

    @property
    def dataframes(self):
        return self._dataframes

    def _load_model(self):
        if self._res_type == "application/x-parquet":
            self._model_data = self._read_parquet()
        else:
            raise Exception("Unknown res_type {self._res_type}")

    def _read_parquet(self):
        s3_key = self._get_key(self._selected_key)
        obj = self._s3_client.get_object(Bucket=self._bucket, Key=s3_key)
        dataframe = pd.read_parquet(BytesIO(obj["Body"].read()))
        return dataframe

    def select(self, filter_dict):
        if self._model_selected:
            raise Exception(
                "Cannot select two different models. Call get_ds_model again to use another model."
            )
        self._model_selected = True
        dataframe = find_matching_model(self._dataframes, filter_dict)
        if dataframe is not None:
            self._selected_key = dataframe["key"]

    def invoke(self):
        self._log.debug("Invoke: Start")
        if not self._model_selected:
            raise Exception("You must call select before invoke")
        if self._selected_key is None:
            return None
        self._load_model()
        return self._model_data



# pylint: disable=missing-docstring
# pylint: disable=unused-import

import pytest
from structlog import get_logger

from .model import create_ds_models


def test_no_models():
    model = create_ds_models(models=[], log=get_logger())
    assert model is not None



from .header_tome import create_subheader_tome_from_fs


class HeaderTomeCopierFs:
    def __init__(
        self, *, src_tome_name, tome_collection_root_path, dest_tome_name, ds_type, log
    ):
        self._src_tome_name = src_tome_name
        self._tome_collection_root_path = tome_collection_root_path
        self._dest_tome_name = dest_tome_name
        self._ds_type = ds_type
        self._log = log

    def copy(self):
        create_subheader_tome_from_fs(
            self._dest_tome_name,
            src_tome_name=self._src_tome_name,
            tome_collection_root_path=self._tome_collection_root_path,
            preserve_src_id=True,
            ds_type=self._ds_type,
            is_copied_header=True,
            log=self._log,
        )



import itertools


def normalize_instructions(instructions):
    final_instructions = []
    instructions = sorted(instructions, key=lambda x: x["channel"])
    channel_dict = {
        key: list(item)
        for key, item in itertools.groupby(instructions, lambda x: x["channel"])
    }
    for channel, channel_instructions in channel_dict.items():
        column_sets = [
            instruction.get("columns") for instruction in channel_instructions
        ]
        if None in column_sets:
            final_instructions.append({"channel": channel})
        else:
            columns_final = []
            for column_set in column_sets:
                columns_final.extend(column_set)
            columns_final = list(
                dict.fromkeys(columns_final)
            )  # remove dupes keeping order
            final_instructions.append({"channel": channel, "columns": columns_final})
    return final_instructions



import pickle

from .find import find_matching_model
from .s3_model import S3Model


class S3ScikitSet(S3Model):
    def __init__(self, *, hdbscan, model, log):
        super().__init__(model, log)
        self._prefix = model["prefix"]
        self._extension = model["extension"]
        self._scikits = model["scikits"]
        self._model_type = model["model_type"]
        self._model_selected = False
        self._selected_key = None
        self._hdbscan = hdbscan
        self._log = log.bind(
            client="s3_scikit_set",
            prefix=self._prefix,
            bucket=self._bucket,
            model_name=self.model_name,
        )

    @property
    def scikits(self):
        return self._scikits

    def _load_model(self):
        if self._res_type == "application/x-pickle":
            model = self._read_pickle()
        else:
            raise Exception(f"Unknown res_type {self._res_type}")
        return model

    def _read_pickle(self):
        s3_key = self._get_key(self._selected_key)
        my_pickle = pickle.loads(
            self._s3_client.get_object(Bucket=self._bucket, Key=s3_key)["Body"].read()
        )
        return my_pickle

    def _use_model(self, model, data):
        if self._model_type == "hdbscan":
            # pylint: disable=unused-variable
            if self._hdbscan is None:
                raise Exception("hdbscan must be injected")
            test_labels, strengths = self._hdbscan.approximate_predict(model, data)
            self._model_data = test_labels
        else:
            raise Exception(f"Unknown model_type {self._model_type}")

    def select(self, filter_dict):
        if self._model_selected:
            raise Exception(
                "Cannot select two different models. Call get_ds_model again to use another model."
            )
        self._model_selected = True
        scikit = find_matching_model(self._scikits, filter_dict)
        if scikit is not None:
            self._selected_key = scikit["key"]

    def invoke(self, dataframe):
        self._log.debug("Invoke: Start")
        if not self._model_selected:
            raise Exception("You must call select before invoke")
        if self._selected_key is None:
            return None
        model = self._load_model()
        self._use_model(model, dataframe)
        return self._model_data



import time

import structlog


class AdxDatasetWriterS3:
    def __init__(self, *, bucket, prefix=None, log=None):
        self._log = log if log is not None else structlog.get_logger()
        self._log = self._log.bind(
            bucket=bucket, prefix=prefix, client="adx_dataset_writer_s3"
        )
        self._bucket = bucket
        self._prefix = prefix

    def export_revision(self, client, dataset_id, revision_id):
        res = client.create_job(
            Type="EXPORT_REVISIONS_TO_S3",
            Details={
                "ExportRevisionsToS3": {
                    "DataSetId": dataset_id,
                    "RevisionDestinations": [
                        {
                            "Bucket": self._bucket,
                            "KeyPattern": self._get_key_pattern(),
                            "RevisionId": revision_id,
                        },
                    ],
                },
            },
        )

        job_id = res["Id"]

        client.start_job(JobId=job_id)

        while True:
            time.sleep(2)
            job = client.get_job(JobId=job_id)

            if job.get("State") == "ERROR":
                message = job.get("Errors")[0].get("Message")
                raise Exception(f"Job {job_id} failed: {message}")

            if job.get("State") == "CANCELLED":
                raise Exception(f"Job {job_id} was cancelled.")

            if job.get("State") == "TIMED_OUT":
                raise Exception(f"Job {job_id} timed out.")

            if job.get("State") == "COMPLETED":
                return

    def auto_export_revisions(self, client, dataset_id):
        action = client.create_event_action(
            Action={
                "ExportRevisionToS3": {
                    "RevisionDestination": {
                        "Bucket": self._bucket,
                        "KeyPattern": self._get_key_pattern(),
                    }
                }
            },
            Event={"RevisionPublished": {"DataSetId": dataset_id}},
        )
        self._log.info("Created Action", event_action_id=action["Id"])
        return action

    def _get_key_pattern(self):
        pattern_prefix = self._prefix if self._prefix is not None else ""
        return "".join([pattern_prefix, "${Asset.Name}"])



# pylint: disable=missing-docstring
# pylint: disable=unused-import

import pytest
from .normalize_instructions import normalize_instructions


def test_combining_instructions():
    instructions = normalize_instructions(
        [
            {"channel": "round_end", "columns": ["tick"]},
            {"channel": "round_end", "columns": ["win_reason_code"]},
        ]
    )
    assert instructions == [
        {"channel": "round_end", "columns": ["tick", "win_reason_code"]}
    ]


def test_combining_instructions_no_col_specified():
    instructions = normalize_instructions(
        [{"channel": "round_end", "columns": ["tick"]}, {"channel": "round_end"}]
    )
    assert instructions == [{"channel": "round_end"}]


def test_combining_three_instructions():
    instructions = normalize_instructions(
        [
            {"channel": "round_end", "columns": ["tick"]},
            {"channel": "round_end", "columns": ["win_reason_code"]},
            {"channel": "round_end", "columns": ["winner_team_code"]},
        ]
    )
    assert instructions == [
        {
            "channel": "round_end",
            "columns": ["tick", "win_reason_code", "winner_team_code"],
        }
    ]


def test_combining_several_instructions():
    instructions = normalize_instructions(
        [
            {"channel": "ch1", "columns": ["col1"]},
            {"channel": "ch1", "columns": ["col2"]},
            {"channel": "ch2", "columns": ["col1"]},
            {"channel": "ch2", "columns": ["col1", "col2"]},
            {"channel": "ch2", "columns": ["col4"]},
            {"channel": "ch2", "columns": ["col1", "col2", "col3", "col4"]},
            {"channel": "ch3", "columns": ["col1"]},
            {"channel": "ch3", "columns": ["col1", "col2"]},
            {"channel": "ch3", "columns": ["col3"]},
            {"channel": "ch3"},
        ]
    )
    assert instructions == [
        {"channel": "ch1", "columns": ["col1", "col2"]},
        {"channel": "ch2", "columns": ["col1", "col2", "col4", "col3"]},
        {"channel": "ch3"},
    ]


def test_combining_out_of_order_instructions():
    instructions = normalize_instructions(
        [
            {"channel": "ch3", "columns": ["col1", "col2"]},
            {"channel": "ch1", "columns": ["col1"]},
            {"channel": "ch2", "columns": ["col1", "col2"]},
            {"channel": "ch2", "columns": ["col1"]},
            {"channel": "ch2", "columns": ["col4"]},
            {"channel": "ch2", "columns": ["col1", "col2", "col3", "col4"]},
            {"channel": "ch3", "columns": ["col1"]},
            {"channel": "ch3", "columns": ["col3"]},
            {"channel": "ch1", "columns": ["col2"]},
            {"channel": "ch3"},
        ]
    )
    assert instructions == [
        {"channel": "ch1", "columns": ["col1", "col2"]},
        {"channel": "ch2", "columns": ["col1", "col2", "col4", "col3"]},
        {"channel": "ch3"},
    ]



import structlog
import pandas as pd


class TomeScribe:
    def __init__(
        self,
        *,
        manifest,
        writer,
        max_page_size_mb=None,
        max_page_row_count=None,
        limit_check_frequency=100,
        log: object = None,
    ):
        self._log = log if log is not None else structlog.get_logger()
        self._manifest = manifest
        self._writer = writer
        self._max_page_size_mb = max_page_size_mb
        self._max_page_row_count = max_page_row_count
        self._limit_check_frequency = limit_check_frequency

        self._data_dict = {}
        self._data_dict_index = 0
        self._keyset = []
        self._data_df = None
        self._page_counter = 0

    @property
    def dataframe(self):
        if self._data_df is None:
            self._data_df = pd.DataFrame.from_dict(self._data_dict, "index")
        return self._data_df

    @property
    def keyset(self):
        return self._keyset

    @property
    def tome_name(self):
        return self._writer.tome_name

    @property
    def page_counter(self):
        return self._page_counter

    @property
    def page_size_mb(self):
        return self._get_page_size_mb()

    @property
    def page_row_count(self):
        return self._get_page_row_count()

    def start(self):
        self._writer.write_manifest(self._manifest.get())
        self._manifest.start_page()

    def finish(self):
        if self._page_counter == 0 and len(self._keyset) == 0:
            raise Exception("Empty Tome not supported")
        if len(self._keyset) > 0:
            self._write()
        self._manifest.finish()
        self._writer.write_manifest(self._manifest.get())

    def concat(self, df, keys):
        self._concat_keys(keys)
        self._concat_df(df)
        self._on_data()

    def set_manifest_data(self, data):
        self._page_counter = len(data["pages"])
        self._manifest.set(data)

    def _write(self):
        page = self._manifest.end_page(self._page_counter)
        self._writer.write_page(page, self.dataframe, self.keyset)
        self._writer.write_manifest(self._manifest.get())
        self._page_counter += 1
        self._new_page()

    def _on_data(self):
        if self._will_write_page():
            self._write()

    def _will_write_page(self) -> bool:
        if len(self.keyset) == 0:
            return False
        if len(self.keyset) % self._limit_check_frequency != 0:
            return False
        if self._max_page_size_mb is not None:
            current_size = self._get_page_size_mb()
            if current_size > self._max_page_size_mb:
                return True
        if self._max_page_row_count is not None:
            current_row_count = self._get_page_row_count()
            if current_row_count > self._max_page_row_count:
                return True
        return False

    def _new_page(self) -> None:
        self._keyset = []
        self._data_dict = {}
        self._data_dict_index = 0
        self._data_df = None
        self._manifest.start_page()

    def _concat_df(self, df):
        if df is not None:
            self._data_df = None
            temp_dict = df.to_dict(orient="records")
            for entry in temp_dict:
                self._data_dict[self._data_dict_index] = entry
                self._data_dict_index += 1

    def _concat_keys(self, keys):
        if isinstance(keys, list):
            self._keyset += keys
        else:
            self._keyset.append(keys)

    def _get_page_size_mb(self) -> float:
        df = self.dataframe
        size_in_mb = sum(df.memory_usage()) / 1024 / 1024
        return size_in_mb

    def _get_page_row_count(self) -> int:
        df = self.dataframe
        row_count = df.shape[0]
        return row_count



import os.path
from gzip import GzipFile
from typing import List, Dict, Optional, Union

import structlog
import rapidjson
import pandas as pd
from .handle_value_error import handle_value_error


class DsReaderFs:
    def __init__(
        self,
        *,
        root_path: str,
        manifest_key: str,
        prefix: Optional[str] = None,
        log: object = None,
    ):
        self._log = log if log is not None else structlog.get_logger()
        self._log = self._log.bind(
            client="ds_reader_fs", root_path=root_path, manifest_key=manifest_key
        )
        self._is_gzipped = True
        self._root_path = root_path
        self._manifest_key = manifest_key
        self._prefix = prefix

    def read_manifest(self) -> Dict:
        self._log.info("Read Manifest: Start")

        file_location = os.path.join(
            self._root_path, add_prefix(self._manifest_key, self._prefix)
        )
        if self._is_gzipped:
            with GzipFile(file_location, "r") as fin:
                data = rapidjson.loads(fin.read().decode("utf-8"))
        else:
            with open(file_location, "r", encoding="utf-8") as fin:
                data = rapidjson.loads(fin.read())

        return data

    def read_metadata(self) -> Dict:
        self._log.info("Read Metadata: Start")
        return {}

    def read_parquet_channel(self, channel: Dict, columns: List[str]) -> pd.DataFrame:
        self._log.info(
            "Read parquet Channel: Start", channel=channel["channel"], columns=columns
        )
        file_location = self._get_file_location(channel)
        try:
            df = pd.read_parquet(file_location, columns=columns)
        except ValueError as value_error:
            df = handle_value_error(value_error, channel)

        return df

    def _get_file_location(self, channel, /) -> str:
        key = self._get_file_path(channel)
        return os.path.join(self._root_path, key)

    def _get_file_path(self, channel: Union[str, Dict]) -> str:
        if isinstance(channel, str):
            key = os.path.join(
                os.path.normpath(self._manifest_key).split(os.path.sep)[:-1]
            )

            key = os.path.join(key, channel)

        elif isinstance(channel, dict):
            key = os.path.normpath(channel["key"])
        else:
            raise Exception(f"Unknown channel type {type(channel)}")
        return add_prefix(key, self._prefix)


def add_prefix(key, prefix, /) -> str:
    if prefix is None:
        return key
    return os.path.join(*[*prefix.split("/"), key])
